@@ -20,16 +20,17 @@ import shutil
 import subprocess
 import sys
 
-from scripts.test.support import run_command, split_wast
+from scripts.test.support import run_command, split_wast, node_test_glue, node_has_webassembly
 from scripts.test.shared import (
-    ASM2WASM, BIN_DIR, EMCC, MOZJS, NATIVECC, NATIVEXX, NODEJS, S2WASM_EXE,
-    WASM_AS, WASM_CTOR_EVAL, WASM_OPT, WASM_SHELL, WASM_MERGE, WASM_SHELL_EXE,
+    BIN_DIR, EMCC, MOZJS, NATIVECC, NATIVEXX, NODEJS, S2WASM_EXE,
+    WASM_AS, WASM_CTOR_EVAL, WASM_OPT, WASM_SHELL, WASM_MERGE, WASM_SHELL_EXE, WASM_METADCE,
     WASM_DIS, WASM_REDUCE, binary_format_check, delete_from_orbit, fail, fail_with_error,
     fail_if_not_identical, fail_if_not_contained, has_vanilla_emcc,
     has_vanilla_llvm, minify_check, num_failures, options, tests,
     requested, warnings, has_shell_timeout
 )
 
+import scripts.test.asm2wasm as asm2wasm
 import scripts.test.s2wasm as s2wasm
 import scripts.test.wasm2asm as wasm2asm
 
@@ -144,6 +145,7 @@ def run_wasm_opt_tests():
       actual = actual.replace('printing before:\n', '')
 
       expected = open(f, 'rb').read()
+
       if actual != expected:
         fail(actual, expected)
 
@@ -152,126 +154,6 @@ def run_wasm_opt_tests():
 
       minify_check(t)
 
-def run_asm2wasm_tests():
-  print '[ checking asm2wasm testcases... ]\n'
-
-  for asm in tests:
-    if asm.endswith('.asm.js'):
-      for precise in [0, 1, 2]:
-        for opts in [1, 0]:
-          cmd = ASM2WASM + [os.path.join(options.binaryen_test, asm)]
-          wasm = asm.replace('.asm.js', '.fromasm')
-          if not precise:
-            cmd += ['--emit-potential-traps', '--ignore-implicit-traps']
-            wasm += '.imprecise'
-          elif precise == 2:
-            cmd += ['--emit-clamped-potential-traps']
-            wasm += '.clamp'
-          if not opts:
-            wasm += '.no-opts'
-            if precise:
-              cmd += ['-O0'] # test that -O0 does nothing
-          else:
-            cmd += ['-O']
-          if 'debugInfo' in asm:
-            cmd += ['-g']
-          if 'noffi' in asm:
-            cmd += ['--no-legalize-javascript-ffi']
-          if precise and opts:
-            # test mem init importing
-            open('a.mem', 'wb').write(asm)
-            cmd += ['--mem-init=a.mem']
-            if asm[0] == 'e':
-              cmd += ['--mem-base=1024']
-          if 'i64' in asm or 'wasm-only' in asm or 'noffi' in asm:
-            cmd += ['--wasm-only']
-          wasm = os.path.join(options.binaryen_test, wasm)
-          print '..', asm, wasm
-
-          def do_asm2wasm_test():
-            actual = run_command(cmd)
-
-            # verify output
-            if not os.path.exists(wasm):
-              fail_with_error('output .wast file %s does not exist' % wasm)
-            expected = open(wasm, 'rb').read()
-            if actual != expected:
-              fail(actual, expected)
-
-            binary_format_check(wasm, verify_final_result=False)
-
-          # test both normally and with pass debug (so each inter-pass state is validated)
-          old_pass_debug = os.environ.get('BINARYEN_PASS_DEBUG')
-          try:
-            os.environ['BINARYEN_PASS_DEBUG'] = '1'
-            do_asm2wasm_test()
-            del os.environ['BINARYEN_PASS_DEBUG']
-            do_asm2wasm_test()
-          finally:
-            if old_pass_debug is not None:
-              os.environ['BINARYEN_PASS_DEBUG'] = old_pass_debug
-            else:
-              if 'BINARYEN_PASS_DEBUG' in os.environ:
-                del os.environ['BINARYEN_PASS_DEBUG']
-
-          # verify in wasm
-          if options.interpreter:
-            # remove imports, spec interpreter doesn't know what to do with them
-            subprocess.check_call(WASM_OPT + ['--remove-imports', wasm], stdout=open('ztemp.wast', 'w'), stderr=subprocess.PIPE)
-            proc = subprocess.Popen([options.interpreter, 'ztemp.wast'], stderr=subprocess.PIPE)
-            out, err = proc.communicate()
-            if proc.returncode != 0:
-              try: # to parse the error
-                reported = err.split(':')[1]
-                start, end = reported.split('-')
-                start_line, start_col = map(int, start.split('.'))
-                lines = open('ztemp.wast').read().split('\n')
-                print
-                print '='*80
-                print lines[start_line-1]
-                print (' '*(start_col-1)) + '^'
-                print (' '*(start_col-2)) + '/_\\'
-                print '='*80
-                print err
-              except Exception, e:
-                fail_with_error('wasm interpreter error: ' + err) # failed to pretty-print
-              fail_with_error('wasm interpreter error')
-
-          # verify debug info
-          if 'debugInfo' in asm:
-            jsmap = 'a.wasm.map'
-            cmd += ['--source-map', jsmap,
-                    '--source-map-url', 'http://example.org/' + jsmap,
-                    '-o', 'a.wasm']
-            run_command(cmd)
-            if not os.path.isfile(jsmap):
-              fail_with_error('Debug info map not created: %s' % jsmap)
-            with open(wasm + '.map', 'rb') as expected:
-              with open(jsmap, 'rb') as actual:
-                fail_if_not_identical(actual.read(), expected.read())
-            with open('a.wasm', 'rb') as binary:
-              url_section_name = bytearray([16]) + bytearray('sourceMappingURL')
-              payload = 'http://example.org/' + jsmap
-              assert len(payload) < 256, 'name too long'
-              url_section_contents = bytearray([len(payload)]) + bytearray(payload)
-              print url_section_name
-              binary_contents = bytearray(binary.read())
-              if url_section_name not in binary_contents:
-                fail_with_error('source map url section not found in binary')
-              if url_section_contents not in binary_contents[binary_contents.index(url_section_name):]:
-                fail_with_error('source map url not found in url section')
-
-
-  print '\n[ checking asm2wasm binary reading/writing... ]\n'
-
-  asmjs = os.path.join(options.binaryen_test, 'hello_world.asm.js')
-  delete_from_orbit('a.wasm')
-  delete_from_orbit('b.wast')
-  run_command(ASM2WASM + [asmjs, '-o', 'a.wasm'])
-  assert open('a.wasm', 'rb').read()[0] == '\0', 'we emit binary by default'
-  run_command(ASM2WASM + [asmjs, '-o', 'b.wast', '-S'])
-  assert open('b.wast', 'rb').read()[0] != '\0', 'we emit text with -S'
-
 def run_wasm_dis_tests():
   print '\n[ checking wasm-dis on provided binaries... ]\n'
 
@@ -312,6 +194,17 @@ def run_wasm_merge_tests():
           with open(out + '.stdout') as f:
             fail_if_not_identical(f.read(), stdout)
 
+def run_crash_tests():
+  print "\n[ checking we don't crash on tricky inputs... ]\n"
+
+  for t in os.listdir(os.path.join('test', 'crash')):
+    if t.endswith(('.wast', '.wasm')):
+      print '..', t
+      t = os.path.join('test', 'crash', t)
+      cmd = WASM_OPT + [t]
+      # expect a parse error to be reported
+      run_command(cmd, expected_err='parse exception:', err_contains=True, expected_status=1)
+
 def run_ctor_eval_tests():
   print '\n[ checking wasm-ctor-eval... ]\n'
 
@@ -327,6 +220,23 @@ def run_ctor_eval_tests():
       with open(out) as f:
         fail_if_not_identical(f.read(), actual)
 
+def run_wasm_metadce_tests():
+  print '\n[ checking wasm-metadce ]\n'
+
+  for t in os.listdir(os.path.join('test', 'metadce')):
+    if t.endswith(('.wast', '.wasm')):
+      print '..', t
+      t = os.path.join('test', 'metadce', t)
+      graph = t + '.graph.txt'
+      cmd = WASM_METADCE + [t, '--graph-file=' + graph, '-o', 'a.wast', '-S']
+      stdout = run_command(cmd)
+      expected = t + '.dced'
+      with open('a.wast') as seen:
+        with open(expected) as correct:
+          fail_if_not_identical(seen.read(), correct.read())
+      with open(expected + '.stdout') as correct:
+        fail_if_not_identical(stdout, correct.read())
+
 def run_wasm_reduce_tests():
   print '\n[ checking wasm-reduce ]\n'
 
@@ -439,20 +349,38 @@ def run_spec_tests():
         check_expected(actual, os.path.join(options.binaryen_test, 'spec', 'expected-output', os.path.basename(wast) + '.log'))
 
 def run_binaryen_js_tests():
+  if not MOZJS and not NODEJS:
+    return
+  node_has_wasm = NODEJS and node_has_webassembly(NODEJS)
+
   print '\n[ checking binaryen.js testcases... ]\n'
 
   for s in sorted(os.listdir(os.path.join(options.binaryen_test, 'binaryen.js'))):
     if not s.endswith('.js'): continue
     print s
     f = open('a.js', 'w')
-    f.write(open(os.path.join(options.binaryen_bin, 'binaryen.js')).read())
-    f.write(open(os.path.join(options.binaryen_test, 'binaryen.js', s)).read())
+    binaryen_js = open(os.path.join(options.binaryen_bin, 'binaryen.js')).read()
+    f.write(binaryen_js)
+    if NODEJS:
+      f.write(node_test_glue())
+    test_path = os.path.join(options.binaryen_test, 'binaryen.js', s)
+    test_src = open(test_path).read()
+    f.write(test_src)
     f.close()
-    cmd = [MOZJS, 'a.js']
-    out = run_command(cmd, stderr=subprocess.STDOUT)
-    expected = open(os.path.join(options.binaryen_test, 'binaryen.js', s + '.txt')).read()
-    if expected not in out:
-      fail(out, expected)
+    def test(engine):
+      cmd = [engine, 'a.js']
+      out = run_command(cmd, stderr=subprocess.STDOUT)
+      expected = open(os.path.join(options.binaryen_test, 'binaryen.js', s + '.txt')).read()
+      if expected not in out:
+        fail(out, expected)
+    # run in all possible shells
+    if MOZJS:
+      test(MOZJS)
+    if NODEJS:
+      if node_has_wasm or not 'WebAssembly.' in test_src:
+        test(NODEJS)
+      else:
+        print 'Skipping ' + test_path + ' because WebAssembly might not be supported'
 
 def run_validator_tests():
   print '\n[ running validation tests... ]\n'
@@ -471,31 +399,45 @@ def run_validator_tests():
 def run_torture_tests():
   print '\n[ checking torture testcases... ]\n'
 
-  unexpected_result_count = 0
+  # torture tests are parallel anyhow, don't create multiple threads in each child
+  old_cores = os.environ.get('BINARYEN_CORES')
+  try:
+    os.environ['BINARYEN_CORES'] = '1'
+
+    unexpected_result_count = 0
+
+    import test.waterfall.src.link_assembly_files as link_assembly_files
+    s2wasm_torture_out = os.path.abspath(os.path.join(options.binaryen_test, 's2wasm-torture-out'))
+    if os.path.isdir(s2wasm_torture_out):
+      shutil.rmtree(s2wasm_torture_out)
+    os.mkdir(s2wasm_torture_out)
+    unexpected_result_count += link_assembly_files.run(
+        linker=os.path.abspath(S2WASM_EXE),
+        files=os.path.abspath(os.path.join(options.binaryen_test, 'torture-s', '*.s')),
+        fails=[os.path.abspath(os.path.join(options.binaryen_test, 's2wasm_known_gcc_test_failures.txt'))],
+        attributes=['O2'],
+        out=s2wasm_torture_out,
+        args=None)
+    assert os.path.isdir(s2wasm_torture_out), 'Expected output directory %s' % s2wasm_torture_out
+
+    import test.waterfall.src.execute_files as execute_files
+    unexpected_result_count += execute_files.run(
+        runner=os.path.abspath(WASM_SHELL_EXE),
+        files=os.path.abspath(os.path.join(s2wasm_torture_out, '*.wast')),
+        fails=[os.path.abspath(os.path.join(options.binaryen_test, 's2wasm_known_binaryen_shell_test_failures.txt'))],
+        attributes=['O2'],
+        out='',
+        wasmjs='')
 
-  import test.waterfall.src.link_assembly_files as link_assembly_files
-  s2wasm_torture_out = os.path.abspath(os.path.join(options.binaryen_test, 's2wasm-torture-out'))
-  if os.path.isdir(s2wasm_torture_out):
     shutil.rmtree(s2wasm_torture_out)
-  os.mkdir(s2wasm_torture_out)
-  unexpected_result_count += link_assembly_files.run(
-      linker=os.path.abspath(S2WASM_EXE),
-      files=os.path.abspath(os.path.join(options.binaryen_test, 'torture-s', '*.s')),
-      fails=os.path.abspath(os.path.join(options.binaryen_test, 's2wasm_known_gcc_test_failures.txt')),
-      out=s2wasm_torture_out)
-  assert os.path.isdir(s2wasm_torture_out), 'Expected output directory %s' % s2wasm_torture_out
+    if unexpected_result_count:
+      fail('%s failures' % unexpected_result_count, '0 failures')
 
-  import test.waterfall.src.execute_files as execute_files
-  unexpected_result_count += execute_files.run(
-      runner=os.path.abspath(WASM_SHELL_EXE),
-      files=os.path.abspath(os.path.join(s2wasm_torture_out, '*.wast')),
-      fails=os.path.abspath(os.path.join(options.binaryen_test, 's2wasm_known_binaryen_shell_test_failures.txt')),
-      out='',
-      wasmjs='')
-
-  shutil.rmtree(s2wasm_torture_out)
-  if unexpected_result_count:
-    fail('%s failures' % unexpected_result_count, '0 failures')
+  finally:
+    if old_cores:
+      os.environ['BINARYEN_CORES'] = old_cores
+    else:
+      del os.environ['BINARYEN_CORES']
 
 def run_vanilla_tests():
   print '\n[ checking emcc WASM_BACKEND testcases...]\n'
@@ -591,23 +533,6 @@ def run_gcc_torture_tests():
         fail(actual, expected)
 
 def run_emscripten_tests():
-  if MOZJS and 0:
-
-    print '\n[ checking native wasm support ]\n'
-
-    command = [EMCC, '-o', 'a.wasm.js', '-s', 'BINARYEN=1', os.path.join(options.binaryen_test, 'hello_world.c'), '-s', 'BINARYEN_METHOD="native-wasm"', '-s', 'BINARYEN_SCRIPTS="spidermonkify.py"']
-    print ' '.join(command)
-    subprocess.check_call(command)
-
-    cmd = [MOZJS, 'a.wasm.js']
-    out = run_command(cmd)
-    assert 'hello, world!' in out, out
-
-    proc = subprocess.Popen([NODEJS, 'a.wasm.js'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-    out, err = proc.communicate()
-    assert proc.returncode != 0, 'should fail on no wasm support'
-    assert 'no native wasm support detected' in err, err
-
   print '\n[ checking wasm.js methods... ]\n'
 
   for method_init in ['interpret-asm2wasm', 'interpret-s-expr', 'asmjs', 'interpret-binary', 'asmjs,interpret-binary', 'interpret-binary,asmjs']:
@@ -661,94 +586,48 @@ def run_emscripten_tests():
           assert proc.returncode != 0, err
           assert 'hello, world!' not in out, out
 
-  print '\n[ checking wasm.js testcases... ]\n'
-
-  for c in tests:
-    if c.endswith(('.c', '.cpp')):
-      print '..', c
-      base = c.replace('.cpp', '').replace('.c', '')
-      post = base + '.post.js'
-      try:
-        post = open(os.path.join(options.binaryen_test, post)).read()
-      except:
-        post = None
-      expected = open(os.path.join(options.binaryen_test, base + '.txt')).read()
-      emcc = os.path.join(options.binaryen_test, base + '.emcc')
-      extra = []
-      if os.path.exists(emcc):
-        extra = json.loads(open(emcc).read())
-      if os.path.exists('a.normal.js'): os.unlink('a.normal.js')
-      for opts in [[], ['-O1'], ['-O2'], ['-O3'], ['-Oz']]:
-        for method in ['interpret-asm2wasm', 'interpret-s-expr', 'interpret-binary']:
-          command = [EMCC, '-o', 'a.wasm.js', '-s', 'BINARYEN=1', os.path.join(options.binaryen_test, c)] + opts + extra
-          command += ['-s', 'BINARYEN_METHOD="' + method + '"']
-          command += ['-s', 'BINARYEN_TRAP_MODE="js"']
-          print '....' + ' '.join(command)
-          subprocess.check_call(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-          if post:
-            open('a.wasm.js', 'a').write(post)
-          else:
-            print '     (no post)'
-          for which in ['wasm']:
-            print '......', which
-            try:
-              args = json.loads(open(os.path.join(options.binaryen_test, base + '.args')).read())
-            except:
-              args = []
-              print '     (no args)'
-
-            def execute():
-              if NODEJS:
-                cmd = [NODEJS, 'a.' + which + '.js'] + args
-                out = run_command(cmd)
-                if out.strip() != expected.strip():
-                  fail(out, expected)
-
-            execute()
-
-            if method in ['interpret-s-expr']:
-              # binary and back
-              shutil.copyfile('a.wasm.wast', 'a.wasm.original.wast')
-              recreated = binary_format_check('a.wasm.wast', verify_final_result=False)
-              shutil.copyfile(recreated, 'a.wasm.wast')
-              execute()
-
-
 # Run all the tests
-run_help_tests()
-run_wasm_opt_tests()
-run_asm2wasm_tests()
-run_wasm_dis_tests()
-run_wasm_merge_tests()
-run_ctor_eval_tests()
-if has_shell_timeout():
-  run_wasm_reduce_tests()
+def main():
+  run_help_tests()
+  run_wasm_opt_tests()
+  asm2wasm.test_asm2wasm()
+  asm2wasm.test_asm2wasm_binary()
+  run_wasm_dis_tests()
+  run_wasm_merge_tests()
+  run_crash_tests()
+  run_ctor_eval_tests()
+  run_wasm_metadce_tests()
+  if has_shell_timeout():
+    run_wasm_reduce_tests()
 
-run_spec_tests()
-if MOZJS:
+  run_spec_tests()
   run_binaryen_js_tests()
-s2wasm.test_s2wasm()
-s2wasm.test_linker()
-wasm2asm.test_wasm2asm()
-run_validator_tests()
-if options.torture and options.test_waterfall:
-  run_torture_tests()
-if has_vanilla_emcc and has_vanilla_llvm and 0:
-  run_vanilla_tests()
-print '\n[ checking example testcases... ]\n'
-if options.run_gcc_tests:
-  run_gcc_torture_tests()
-if EMCC:
-  run_emscripten_tests()
+  s2wasm.test_s2wasm()
+  s2wasm.test_linker()
+  wasm2asm.test_wasm2asm()
+  run_validator_tests()
+  if options.torture and options.test_waterfall:
+    run_torture_tests()
+  if has_vanilla_emcc and has_vanilla_llvm and 0:
+    run_vanilla_tests()
+  print '\n[ checking example testcases... ]\n'
+  if options.run_gcc_tests:
+    run_gcc_torture_tests()
+  if EMCC:
+    run_emscripten_tests()
 
-# Check/display the results
-if num_failures == 0:
-  print '\n[ success! ]'
+  # Check/display the results
+  if num_failures == 0:
+    print '\n[ success! ]'
 
-if warnings:
-  print '\n' + '\n'.join(warnings)
+  if warnings:
+    print '\n' + '\n'.join(warnings)
 
-if num_failures > 0:
-  print '\n[ ' + str(num_failures) + ' failures! ]'
+  if num_failures > 0:
+    print '\n[ ' + str(num_failures) + ' failures! ]'
+
+  sys.exit(num_failures)
+
+if __name__ == '__main__':
+  main()
 
-sys.exit(num_failures)
