@@ -25,9 +25,9 @@ parser.add_argument(
     help='If enabled, only fetches the waterfall build. Default: false.')
 parser.add_argument(
     '--test-waterfall', dest='test_waterfall', action='store_true',
-    default=True,
+    default=False,
     help=('If enabled, fetches and tests the LLVM waterfall builds.'
-          ' Default: true.'))
+          ' Default: false.'))
 parser.add_argument(
     '--no-test-waterfall', dest='test_waterfall', action='store_false',
     help='Disables downloading and testing of the LLVM waterfall builds.')
@@ -102,6 +102,10 @@ if not options.binaryen_bin:
   else:
     options.binaryen_bin = 'bin'
 
+# ensure BINARYEN_ROOT is set up
+os.environ['BINARYEN_ROOT'] = os.path.dirname(os.path.abspath(
+    options.binaryen_bin))
+
 options.binaryen_bin = os.path.normpath(options.binaryen_bin)
 
 wasm_dis_filenames = ['wasm-dis', 'wasm-dis.exe']
@@ -164,6 +168,7 @@ WASM_SHELL = [os.path.join(options.binaryen_bin, 'wasm-shell')]
 WASM_MERGE = [os.path.join(options.binaryen_bin, 'wasm-merge')]
 S2WASM = [os.path.join(options.binaryen_bin, 's2wasm')]
 WASM_REDUCE = [os.path.join(options.binaryen_bin, 'wasm-reduce')]
+WASM_METADCE = [os.path.join(options.binaryen_bin, 'wasm-metadce')]
 
 S2WASM_EXE = S2WASM[0]
 WASM_SHELL_EXE = WASM_SHELL[0]
@@ -240,7 +245,7 @@ def setup_waterfall():
     subprocess.check_call([CLANG, '-v'])
     has_vanilla_llvm = True
     print '...success'
-  except Exception, e:
+  except (OSError, subprocess.CalledProcessError) as e:
     warn('could not run vanilla LLVM from waterfall: ' + str(e) +
          ', looked for clang at ' + CLANG)
 
@@ -256,25 +261,31 @@ if options.only_prepare:
 # external tools
 
 try:
-  subprocess.check_call(
-      [NODEJS, '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-except:
+  if NODEJS is not None:
+    subprocess.check_call(
+        [NODEJS, '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+except (OSError, subprocess.CalledProcessError):
   NODEJS = None
+if NODEJS is None:
   warn('no node found (did not check proper js form)')
 
 try:
-  subprocess.check_call(
-      [MOZJS, '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-except:
+  if MOZJS is not None:
+    subprocess.check_call(
+        [MOZJS, '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+except (OSError, subprocess.CalledProcessError):
   MOZJS = None
+if MOZJS is None:
   warn('no mozjs found (did not check native wasm support nor asm.js'
        ' validation)')
 
 try:
-  subprocess.check_call(
-      [EMCC, '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-except:
+  if EMCC is not None:
+    subprocess.check_call(
+        [EMCC, '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+except (OSError, subprocess.CalledProcessError):
   EMCC = None
+if EMCC is None:
   warn('no emcc found (did not check non-vanilla emscripten/binaryen'
        ' integration)')
 
@@ -284,7 +295,7 @@ try:
       [os.path.join(options.binaryen_test, 'emscripten', 'emcc'), '--version'],
       stdout=subprocess.PIPE, stderr=subprocess.PIPE)
   has_vanilla_emcc = True
-except:
+except (OSError, subprocess.CalledProcessError):
   pass
 
 
@@ -294,13 +305,13 @@ except:
 def delete_from_orbit(filename):
   try:
     os.unlink(filename)
-  except:
+  except OSError:
     pass
   if not os.path.exists(filename):
     return
   try:
     shutil.rmtree(filename, ignore_errors=True)
-  except:
+  except OSError:
     pass
   if not os.path.exists(filename):
     return
@@ -315,7 +326,7 @@ def delete_from_orbit(filename):
       else:
         raise
     shutil.rmtree(filename, onerror=remove_readonly_and_try_again)
-  except:
+  except OSError:
     pass
 
 
