@@ -62,6 +62,7 @@ struct PassOptions {
   int shrinkLevel = 0;   // 0, 1, 2 correspond to -O0, -Os, -Oz
   bool ignoreImplicitTraps = false; // optimize assuming things like div by 0, bad load/store, will not trap
   bool debugInfo = false; // whether to try to preserve debug info through, which are special calls
+  FeatureSet features = Feature::MVP; // Which wasm features to accept, and be allowed to use
 };
 
 //
@@ -76,6 +77,10 @@ struct PassRunner {
   PassRunner(Module* wasm) : wasm(wasm), allocator(&wasm->allocator) {}
   PassRunner(Module* wasm, PassOptions options) : wasm(wasm), allocator(&wasm->allocator), options(options) {}
 
+  // no copying, we control |passes|
+  PassRunner(const PassRunner&) = delete;
+  PassRunner& operator=(const PassRunner&) = delete;
+
   void setDebug(bool debug_) {
     options.debug = debug_;
     options.validateGlobally = debug_; // validate everything by default if debugging
@@ -83,6 +88,9 @@ struct PassRunner {
   void setValidateGlobally(bool validate) {
     options.validateGlobally = validate;
   }
+  void setFeatures(FeatureSet features) {
+    options.features = features;
+  }
 
   void add(std::string passName) {
     auto pass = PassRegistry::get()->createPass(passName);
@@ -122,7 +130,7 @@ struct PassRunner {
   void run();
 
   // Run the passes on a specific function
-  void runFunction(Function* func);
+  void runOnFunction(Function* func);
 
   // Get the last pass that was already executed of a certain type.
   template<class P>
@@ -174,7 +182,7 @@ public:
 
   // Implement this with code to run the pass on a single function, for
   // a function-parallel pass
-  virtual void runFunction(PassRunner* runner, Module* module, Function* function) {
+  virtual void runOnFunction(PassRunner* runner, Module* module, Function* function) {
     WASM_UNREACHABLE();
   }
 
@@ -213,6 +221,9 @@ template <typename WalkerType>
 class WalkerPass : public Pass, public WalkerType {
   PassRunner *runner;
 
+protected:
+  typedef WalkerPass<WalkerType> super;
+
 public:
   void run(PassRunner* runner, Module* module) override {
     setPassRunner(runner);
@@ -220,7 +231,7 @@ public:
     WalkerType::walkModule(module);
   }
 
-  void runFunction(PassRunner* runner, Module* module, Function* func) override {
+  void runOnFunction(PassRunner* runner, Module* module, Function* func) override {
     setPassRunner(runner);
     WalkerType::setModule(module);
     WalkerType::walkFunction(func);
