@@ -59,9 +59,10 @@
 #include "wasm.h"
 #include "pass.h"
 #include "wasm-builder.h"
-#include "ast_utils.h"
-#include "ast/branch-utils.h"
-#include "ast/label-utils.h"
+#include "ir/utils.h"
+#include "ir/branch-utils.h"
+#include "ir/effects.h"
+#include "ir/label-utils.h"
 
 namespace wasm {
 
@@ -126,9 +127,12 @@ struct CodeFolding : public WalkerPass<ControlFlowWalker<CodeFolding>> {
     if (curr->condition || curr->value) {
       unoptimizables.insert(curr->name);
     } else {
-      // we can only optimize if we are at the end of the parent block
+      // we can only optimize if we are at the end of the parent block,
+      // and if the parent block does not return a value (we can't move
+      // elements out of it if there is a value being returned)
       Block* parent = controlFlowStack.back()->dynCast<Block>();
-      if (parent && curr == parent->list.back()) {
+      if (parent && curr == parent->list.back() &&
+          !isConcreteWasmType(parent->list.back()->type)) {
         breakTails[curr->name].push_back(Tail(curr, parent));
       } else {
         unoptimizables.insert(curr->name);
@@ -167,8 +171,13 @@ struct CodeFolding : public WalkerPass<ControlFlowWalker<CodeFolding>> {
   }
 
   void visitBlock(Block* curr) {
+    if (curr->list.empty()) return;
     if (!curr->name.is()) return;
     if (unoptimizables.count(curr->name) > 0) return;
+    // we can't optimize a fallthrough value
+    if (isConcreteWasmType(curr->list.back()->type)) {
+      return;
+    }
     auto iter = breakTails.find(curr->name);
     if (iter == breakTails.end()) return;
     // looks promising
@@ -221,7 +230,7 @@ struct CodeFolding : public WalkerPass<ControlFlowWalker<CodeFolding>> {
     anotherPass = true;
     while (anotherPass) {
       anotherPass = false;
-      WalkerPass<ControlFlowWalker<CodeFolding>>::doWalkFunction(func);
+      super::doWalkFunction(func);
       optimizeTerminatingTails(unreachableTails);
       // optimize returns at the end, so we can benefit from a fallthrough if there is a value TODO: separate passes for them?
       optimizeTerminatingTails(returnTails);
@@ -375,8 +384,14 @@ private:
       if (!tail.isFallthrough()) {
         tail.block->list.push_back(last);
       }
-      // the block type may change if we removed final values
-      tail.block->finalize();
+      // the block type may change if we removed unreachable stuff,
+      // but in general it should remain the same, as if it had a
+      // forced type it should remain, *and*, we don't have a
+      // fallthrough value (we would never get here), so a concrete
+      // type was not from that. I.e., any type on the block is
+      // either forced and/or from breaks with a value, so the
+      // type cannot be changed by moving code out.
+      tail.block->finalize(tail.block->type);
     }
     // since we managed a merge, then it might open up more opportunities later
     anotherPass = true;
@@ -474,9 +489,18 @@ private:
     };
     // let's see if we can merge deeper than num, to num + 1
     auto next = tails;
-    // remove tails that are too short
+    // remove tails that are too short, or that we hit an item we can't handle
     next.erase(std::remove_if(next.begin(), next.end(), [&](Tail& tail) {
-      return effectiveSize(tail) < num + 1;
+      if (effectiveSize(tail) < num + 1) return true;
+      auto* newItem = getItem(tail, num);
+      // ignore tails that break to outside blocks. we want to move code to
+      // the very outermost position, so such code cannot be moved
+      // TODO: this should not be a problem in *non*-terminating tails,
+      //       but double-verify that
+      if (EffectAnalyzer(getPassOptions(), newItem).hasExternalBreakTargets()) {
+        return true;
+      }
+      return false;
     }), next.end());
     // if we have enough to investigate, do so
     if (next.size() >= 2) {
