@@ -24,12 +24,13 @@
 #include <pass.h>
 #include <wasm-s-parser.h>
 #include <support/threads.h>
-#include <ast_utils.h>
-#include <ast/cost.h>
-#include <ast/effects.h>
-#include <ast/manipulation.h>
-#include <ast/properties.h>
-#include <ast/literal-utils.h>
+#include <ir/utils.h>
+#include <ir/cost.h>
+#include <ir/effects.h>
+#include <ir/manipulation.h>
+#include <ir/properties.h>
+#include <ir/literal-utils.h>
+#include <ir/load-utils.h>
 
 // TODO: Use the new sign-extension opcodes where appropriate. This needs to be conditionalized on the availability of atomics.
 
@@ -243,7 +244,8 @@ Index getMaxBits(Expression* curr, LocalInfoProvider* localInfoProvider) {
     return localInfoProvider->getMaxBitsForLocal(get);
   } else if (auto* load = curr->dynCast<Load>()) {
     // if signed, then the sign-extension might fill all the bits
-    if (!load->signed_) {
+    // if unsigned, then we have a limit
+    if (LoadUtils::isSignRelevant(load) && !load->signed_) {
       return 8 * load->bytes;
     }
   }
@@ -320,7 +322,7 @@ struct LocalScanner : PostWalker<LocalScanner> {
     if (Properties::getSignExtValue(value)) {
       signExtBits = Properties::getSignExtBits(value);
     } else if (auto* load = value->dynCast<Load>()) {
-      if (load->signed_) {
+      if (LoadUtils::isSignRelevant(load) && load->signed_) {
         signExtBits = load->bytes * 8;
       }
     }
@@ -364,7 +366,7 @@ struct OptimizeInstructions : public WalkerPass<PostWalker<OptimizeInstructions,
       scanner.walkFunction(func);
     }
     // main walk
-    WalkerPass<PostWalker<OptimizeInstructions, UnifiedExpressionVisitor<OptimizeInstructions>>>::doWalkFunction(func);
+    super::doWalkFunction(func);
   }
 
   void visitExpression(Expression* curr) {
@@ -420,7 +422,8 @@ struct OptimizeInstructions : public WalkerPass<PostWalker<OptimizeInstructions,
         if (extraShifts == 0) {
           if (auto* load = getFallthrough(ext)->dynCast<Load>()) {
             // pattern match a load of 8 bits and a sign extend using a shl of 24 then shr_s of 24 as well, etc.
-            if ((load->bytes == 1 && bits == 8) || (load->bytes == 2 && bits == 16)) {
+            if (LoadUtils::canBeSigned(load) &&
+                ((load->bytes == 1 && bits == 8) || (load->bytes == 2 && bits == 16))) {
               // if the value falls through, we can't alter the load, as it might be captured in a tee
               if (load->signed_ == true || load == ext) {
                 load->signed_ = true;
@@ -559,6 +562,32 @@ struct OptimizeInstructions : public WalkerPass<PostWalker<OptimizeInstructions,
           }
         }
       }
+      // bitwise operations
+      if (binary->op == AndInt32) {
+        // try de-morgan's AND law,
+        //  (eqz X) and (eqz Y) === eqz (X or Y)
+        // Note that the OR and XOR laws do not work here, as these
+        // are not booleans (we could check if they are, but a boolean
+        // would already optimize with the eqz anyhow, unless propagating).
+        // But for AND, the left is true iff X and Y are each all zero bits,
+        // and the right is true if the union of their bits is zero; same.
+        if (auto* left = binary->left->dynCast<Unary>()) {
+          if (left->op == EqZInt32) {
+            if (auto* right = binary->right->dynCast<Unary>()) {
+              if (right->op == EqZInt32) {
+                // reuse one unary, drop the other
+                auto* leftValue = left->value;
+                left->value = binary;
+                binary->left = leftValue;
+                binary->right = right->value;
+                binary->op = OrInt32;
+                return left;
+              }
+            }
+          }
+        }
+      }
+      // for and and or, we can potentially conditionalize
       if (binary->op == AndInt32 || binary->op == OrInt32) {
         return conditionalizeExpensiveOnBitwise(binary);
       }
@@ -862,9 +891,10 @@ private:
             return;
           }
         } else if (curr->op == ShlInt32) {
-          // shifting a 0 is a 0, unless the shift has side effects
-          if (left && Bits::getEffectiveShifts(left) == 0 && !EffectAnalyzer(passOptions, curr->right).hasSideEffects()) {
-            replaceCurrent(left);
+          // shifting a 0 is a 0, or anything by 0 has no effect, all unless the shift has side effects
+          if (((left && left->value.geti32() == 0) || (right && Bits::getEffectiveShifts(right) == 0)) &&
+              !EffectAnalyzer(passOptions, curr->right).hasSideEffects()) {
+            replaceCurrent(curr->left);
             return;
           }
         } else if (curr->op == MulInt32) {
