@@ -22,6 +22,7 @@
 #include <wasm-printing.h>
 #include <pass.h>
 #include <pretty_printing.h>
+#include <ir/module-utils.h>
 
 namespace wasm {
 
@@ -47,6 +48,8 @@ struct PrintSExpression : public Visitor<PrintSExpression> {
   Function* currFunction = nullptr;
   Function::DebugLocation lastPrintedLocation;
 
+  std::unordered_map<Name, Index> functionIndexes;
+
   PrintSExpression(std::ostream& o) : o(o) {
     setMinify(false);
     if (!full) full = isFullForced();
@@ -261,7 +264,7 @@ struct PrintSExpression : public Visitor<PrintSExpression> {
     printCallBody(curr);
   }
   void visitCallIndirect(CallIndirect *curr) {
-    printOpening(o, "call_indirect ") << curr->fullType;
+    printOpening(o, "call_indirect (type ") << curr->fullType << ')';
     incIndent();
     for (auto operand : curr->operands) {
       printFullLine(operand);
@@ -352,7 +355,9 @@ struct PrintSExpression : public Visitor<PrintSExpression> {
   }
   static void printRMWSize(std::ostream& o, WasmType type, uint8_t bytes) {
     prepareColor(o) << printWasmType(type) << ".atomic.rmw";
-    if (bytes != getWasmTypeSize(type)) {
+    if (type == unreachable) {
+      o << '?';
+    } else if (bytes != getWasmTypeSize(type)) {
       if (bytes == 1) {
         o << '8';
       } else if (bytes == 2) {
@@ -684,6 +689,14 @@ struct PrintSExpression : public Visitor<PrintSExpression> {
     lastPrintedLocation = { 0, 0, 0 };
     printOpening(o, "func ", true);
     printName(curr->name);
+    if (currModule && !minify) {
+      // emit the function index in a comment
+      if (functionIndexes.empty()) {
+        ModuleUtils::BinaryIndexes indexes(*currModule);
+        functionIndexes = std::move(indexes.functionIndexes);
+      }
+      o << " (; " << functionIndexes[curr->name] << " ;)";
+    }
     if (curr->type.is()) {
       o << maybeSpace << "(type " << curr->type << ')';
     }
@@ -718,40 +731,41 @@ struct PrintSExpression : public Visitor<PrintSExpression> {
   void printTableHeader(Table* curr) {
     printOpening(o, "table") << ' ';
     o << curr->initial;
-    if (curr->max != Table::kMaxSize) o << ' ' << curr->max;
+    if (curr->hasMax()) o << ' ' << curr->max;
     o << " anyfunc)";
   }
   void visitTable(Table *curr) {
+    if (!curr->exists) return;
     // if table wasn't imported, declare it
     if (!curr->imported) {
       doIndent(o, indent);
       printTableHeader(curr);
       o << maybeNewLine;
     }
-    if (curr->segments.empty()) return;
-    doIndent(o, indent);
     for (auto& segment : curr->segments) {
       // Don't print empty segments
       if (segment.data.empty()) continue;
+      doIndent(o, indent);
       printOpening(o, "elem ", true);
       visit(segment.offset);
       for (auto name : segment.data) {
         o << ' ';
         printName(name);
       }
-      o << ')';
+      o << ")\n";
     }
-    o << maybeNewLine;
   }
   void printMemoryHeader(Memory* curr) {
     printOpening(o, "memory") << ' ';
     printName(curr->name) << ' ';
+    if (curr->shared) printOpening(o, "shared ");
     o << curr->initial;
-    if (curr->max && curr->max != Memory::kMaxSize) o << ' ' << curr->max;
-    if (curr->shared) o << " shared";
+    if (curr->hasMax()) o << ' ' << curr->max;
+    if (curr->shared) o << ")";
     o << ")";
   }
   void visitMemory(Memory* curr) {
+    if (!curr->exists) return;
     // if memory wasn't imported, declare it
     if (!curr->imported) {
       doIndent(o, indent);
