@@ -33,7 +33,7 @@
 #include "shell-interface.h"
 #include "optimization-options.h"
 #include "execution-results.h"
-#include "translate-to-fuzz.h"
+#include "fuzzing.h"
 #include "js-wrapper.h"
 #include "spec-wrapper.h"
 
@@ -62,13 +62,13 @@ std::string runCommand(std::string command) {
 
 int main(int argc, const char* argv[]) {
   Name entry;
-  std::vector<std::string> passes;
   bool emitBinary = true;
   bool debugInfo = false;
   bool fuzzExec = false;
   bool fuzzBinary = false;
   std::string extraFuzzCommand;
   bool translateToFuzz = false;
+  bool fuzzPasses = false;
   std::string emitJSWrapper;
   std::string emitSpecWrapper;
 
@@ -98,6 +98,9 @@ int main(int argc, const char* argv[]) {
       .add("--translate-to-fuzz", "-ttf", "Translate the input into a valid wasm module *somehow*, useful for fuzzing",
            Options::Arguments::Zero,
            [&](Options *o, const std::string &arguments) { translateToFuzz = true; })
+      .add("--fuzz-passes", "-fp", "Pick a random set of passes to run, useful for fuzzing. this depends on translate-to-fuzz (it picks the passes from the input)",
+           Options::Arguments::Zero,
+           [&](Options *o, const std::string &arguments) { fuzzPasses = true; })
       .add("--emit-js-wrapper", "-ejw", "Emit a JavaScript wrapper file that can run the wasm with some test values, useful for fuzzing",
            Options::Arguments::One,
            [&](Options *o, const std::string &arguments) { emitJSWrapper = arguments; })
@@ -111,6 +114,9 @@ int main(int argc, const char* argv[]) {
   options.parse(argc, argv);
 
   Module wasm;
+  // It should be safe to just always enable atomics in wasm-opt, because we
+  // don't expect any passes to accidentally generate atomic ops
+  FeatureSet features = Feature::Atomics;
 
   if (options.debug) std::cerr << "reading...\n";
 
@@ -126,14 +132,19 @@ int main(int argc, const char* argv[]) {
       Fatal() << "error in building module, std::bad_alloc (possibly invalid request for silly amounts of memory)";
     }
 
-    if (!WasmValidator().validate(wasm)) {
+    if (!WasmValidator().validate(wasm, features)) {
+      WasmPrinter::printModule(&wasm);
       Fatal() << "error in validating input";
     }
   } else {
     // translate-to-fuzz
-    TranslateToFuzzReader reader(wasm);
-    reader.read(options.extra["infile"]);
-    if (!WasmValidator().validate(wasm)) {
+    TranslateToFuzzReader reader(wasm, options.extra["infile"]);
+    if (fuzzPasses) {
+      reader.pickPasses(options);
+    }
+    reader.build();
+    if (!WasmValidator().validate(wasm, features)) {
+      WasmPrinter::printModule(&wasm);
       std::cerr << "translate-to-fuzz must always generate a valid module";
       abort();
     }
@@ -171,29 +182,38 @@ int main(int argc, const char* argv[]) {
     std::cout << "[extra-fuzz-command first output:]\n" << firstOutput << '\n';
   }
 
+  Module* curr = &wasm;
+  Module other;
+
+  if (fuzzExec && fuzzBinary) {
+    BufferWithRandomAccess buffer(false);
+    // write the binary
+    WasmBinaryWriter writer(&wasm, buffer, false);
+    writer.write();
+    // read the binary
+    auto input = buffer.getAsChars();
+    WasmBinaryBuilder parser(other, input, false);
+    parser.read();
+    bool valid = WasmValidator().validate(other, features);
+    if (!valid) {
+      WasmPrinter::printModule(&other);
+    }
+    assert(valid);
+    curr = &other;
+  }
+
   if (options.runningPasses()) {
     if (options.debug) std::cerr << "running passes...\n";
-    PassRunner passRunner = options.getPassRunner(wasm);
-    passRunner.run();
-    assert(WasmValidator().validate(wasm));
+    options.runPasses(*curr);
+    bool valid = WasmValidator().validate(*curr, features);
+    if (!valid) {
+      WasmPrinter::printModule(&*curr);
+    }
+    assert(valid);
   }
 
   if (fuzzExec) {
-    auto* compare = &wasm;
-    Module second;
-    if (fuzzBinary) {
-      compare = &second;
-      BufferWithRandomAccess buffer(false);
-      // write the binary
-      WasmBinaryWriter writer(&wasm, buffer, false);
-      writer.write();
-      // read the binary
-      auto input = buffer.getAsChars();
-      WasmBinaryBuilder parser(second, input, false);
-      parser.read();
-      assert(WasmValidator().validate(second));
-    }
-    results.check(*compare);
+    results.check(*curr);
   }
 
   if (options.extra.count("output") > 0) {
@@ -202,7 +222,7 @@ int main(int argc, const char* argv[]) {
     writer.setDebug(options.debug);
     writer.setBinary(emitBinary);
     writer.setDebugInfo(debugInfo);
-    writer.write(wasm, options.extra["output"]);
+    writer.write(*curr, options.extra["output"]);
 
     if (extraFuzzCommand.size() > 0) {
       auto secondOutput = runCommand(extraFuzzCommand);
