@@ -30,13 +30,17 @@
 #include "pass.h"
 #include "support/command-line.h"
 #include "support/file.h"
+#include "support/path.h"
 #include "wasm-io.h"
 #include "wasm-builder.h"
-#include "ast/literal-utils.h"
+#include "ir/literal-utils.h"
 #include "wasm-validator.h"
 
 using namespace wasm;
 
+// a timeout on every execution of the command
+size_t timeout = 2;
+
 struct ProgramResult {
   int code;
   std::string output;
@@ -51,10 +55,10 @@ struct ProgramResult {
   void getFromExecution(std::string command) {
     // do this using just core stdio.h and stdlib.h, for portability
     // sadly this requires two invokes
-    code = system(("timeout 2s " + command + " > /dev/null 2> /dev/null").c_str());
+    code = system(("timeout " + std::to_string(timeout) + "s " + command + " > /dev/null 2> /dev/null").c_str());
     const int MAX_BUFFER = 1024;
     char buffer[MAX_BUFFER];
-    FILE *stream = popen(("timeout 2s " + command + " 2> /dev/null").c_str(), "r");
+    FILE *stream = popen(("timeout " + std::to_string(timeout) + "s " + command + " 2> /dev/null").c_str(), "r");
     while (fgets(buffer, MAX_BUFFER, stream) != NULL) {
       output.append(buffer);
     }
@@ -134,7 +138,7 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
       // try both combining with a generic shrink (so minor pass overhead is compensated for), and without
       for (auto shrinking : { false, true }) {
         for (auto pass : passes) {
-          std::string currCommand = "bin/wasm-opt ";
+          std::string currCommand = Path::getBinaryenBinaryTool("wasm-opt") + " ";
           if (shrinking) currCommand += " --dce --vacuum ";
           currCommand += working + " -o " + test + " " + pass;
           if (debugInfo) currCommand += " -g ";
@@ -414,9 +418,8 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
     }
     for (auto& func : functions) {
       curr->removeFunction(func.name);
-      WasmValidator validator;
-      validator.quiet = true;
-      if (validator.validate(*curr) && writeAndTestReduction()) {
+      if (WasmValidator().validate(*curr, Feature::MVP, WasmValidator::Globally | WasmValidator::Quiet) &&
+          writeAndTestReduction()) {
         std::cerr << "|      removed function " << func.name << '\n';
         noteReduction();
       } else {
@@ -527,6 +530,12 @@ int main(int argc, const char* argv[]) {
            [&](Options* o, const std::string& argument) {
              force = true;
            })
+      .add("--timeout", "-to", "A timeout to apply to each execution of the command, in seconds (default: 2)",
+           Options::Arguments::One,
+           [&](Options* o, const std::string& argument) {
+             timeout = atoi(argument.c_str());
+             std::cout << "|applying timeout: " << timeout << "\n";
+           })
       .add_positional("INFILE", Options::Arguments::One,
                       [&](Options* o, const std::string& argument) {
                         input = argument;
@@ -568,7 +577,7 @@ int main(int argc, const char* argv[]) {
   std::cerr << "|checking that command has expected behavior on canonicalized (read-written) binary\n";
   {
     // read and write it
-    ProgramResult readWrite(std::string("bin/wasm-opt ") + input + " -o " + test);
+    ProgramResult readWrite(Path::getBinaryenBinaryTool("wasm-opt") + " " + input + " -o " + test);
     if (readWrite.failed()) {
       stopIfNotForced("failed to read and write the binary", readWrite);
     } else {
