@@ -37,7 +37,14 @@ namespace wasm {
 
 enum {
   // the maximum amount of bytes we emit per LEB
-  MaxLEB32Bytes = 5
+  MaxLEB32Bytes = 5,
+};
+
+// wasm VMs on the web have decided to impose some limits on what they
+// accept
+enum WebLimitations {
+  MaxDataSegments = 100 * 1000,
+  MaxFunctionBodySize = 128 * 1024
 };
 
 template<typename T, typename MiniT>
@@ -659,10 +666,21 @@ class WasmBinaryWriter : public Visitor<WasmBinaryWriter, void> {
 
   void prepare();
 public:
-  WasmBinaryWriter(Module* input, BufferWithRandomAccess& o, bool debug) : wasm(input), o(o), debug(debug) {
+  WasmBinaryWriter(Module* input, BufferWithRandomAccess& o, bool debug = false) : wasm(input), o(o), debug(debug) {
     prepare();
   }
 
+  // locations in the output binary for the various parts of the module
+  struct TableOfContents {
+    struct Entry {
+      Name name;
+      size_t offset; // where the entry starts
+      size_t size; // the size of the entry
+      Entry(Name name, size_t offset, size_t size) : name(name), offset(offset), size(size) {}
+    };
+    std::vector<Entry> functionBodies;
+  } tableOfContents;
+
   void setNamesSection(bool set) { debugInfo = set; }
   void setSourceMap(std::ostream* set, std::string url) {
     sourceMap = set;
@@ -696,8 +714,8 @@ public:
   void writeExports();
   void writeDataSegments();
 
-  std::map<Name, Index> mappedFunctions; // name of the Function => index. first imports, then internals
-  std::map<Name, uint32_t> mappedGlobals; // name of the Global => index. first imported globals, then internal globals
+  std::unordered_map<Name, Index> mappedFunctions; // name of the Function => index. first imports, then internals
+  std::unordered_map<Name, uint32_t> mappedGlobals; // name of the Global => index. first imported globals, then internal globals
   uint32_t getFunctionIndex(Name name);
   uint32_t getGlobalIndex(Name name);
 
@@ -826,8 +844,6 @@ public:
   void readMemory();
   void readSignatures();
 
-  std::vector<Name> functionImportIndexes; // index in function index space => name of function import
-
   // gets a name in the combined function import+defined function space
   Name getFunctionIndexName(Index i);
   void getResizableLimits(Address& initial, Address& max, bool& shared, Address defaultIfNoMax);
@@ -844,7 +860,9 @@ public:
 
   // We read functions before we know their names, so we need to backpatch the names later
   std::vector<Function*> functions; // we store functions here before wasm.addFunction after we know their names
+  std::vector<Import*> functionImports; // we store function imports here before wasm.addFunctionImport after we know their names
   std::map<Index, std::vector<Call*>> functionCalls; // at index i we have all calls to the defined function i
+  std::map<Index, std::vector<CallImport*>> functionImportCalls; // at index i we have all callImports to the imported function i
   Function* currFunction = nullptr;
   Index endOfFunction = -1; // before we see a function (like global init expressions), there is no end of function to check
 
@@ -863,7 +881,7 @@ public:
     BreakTarget(Name name, int arity) : name(name), arity(arity) {}
   };
   std::vector<BreakTarget> breakStack;
-  bool breaksToReturn; // whether a break is done to the function scope, which is in effect a return
+  std::unordered_set<Name> breakTargetNames;
 
   std::vector<Expression*> expressionStack;
 
