@@ -18,7 +18,7 @@
 #define wasm_wasm_builder_h
 
 #include "wasm.h"
-#include "ast/manipulation.h"
+#include "ir/manipulation.h"
 
 namespace wasm {
 
@@ -54,13 +54,15 @@ public:
 
     for (auto& param : params) {
       func->params.push_back(param.type);
-      func->localIndices[param.name] = func->localNames.size();
-      func->localNames.push_back(param.name);
+      Index index = func->localNames.size();
+      func->localIndices[param.name] = index;
+      func->localNames[index] = param.name;
     }
     for (auto& var : vars) {
       func->vars.push_back(var.type);
-      func->localIndices[var.name] = func->localNames.size();
-      func->localNames.push_back(var.name);
+      Index index = func->localNames.size();
+      func->localIndices[var.name] = index;
+      func->localNames[index] = var.name;
     }
 
     return func;
@@ -80,6 +82,7 @@ public:
   Block* makeBlock(Name name, Expression* first = nullptr) {
     auto* ret = makeBlock(first);
     ret->name = name;
+    ret->finalize();
     return ret;
   }
   Block* makeBlock(const std::vector<Expression*>& items) {
@@ -88,12 +91,44 @@ public:
     ret->finalize();
     return ret;
   }
+  Block* makeBlock(const ExpressionList& items) {
+    auto* ret = allocator.alloc<Block>();
+    ret->list.set(items);
+    ret->finalize();
+    return ret;
+  }
+  Block* makeBlock(const ExpressionList& items, WasmType type) {
+    auto* ret = allocator.alloc<Block>();
+    ret->list.set(items);
+    ret->finalize(type);
+    return ret;
+  }
+  Block* makeBlock(Name name, const ExpressionList& items) {
+    auto* ret = allocator.alloc<Block>();
+    ret->name = name;
+    ret->list.set(items);
+    ret->finalize();
+    return ret;
+  }
+  Block* makeBlock(Name name, const ExpressionList& items, WasmType type) {
+    auto* ret = allocator.alloc<Block>();
+    ret->name = name;
+    ret->list.set(items);
+    ret->finalize(type);
+    return ret;
+  }
   If* makeIf(Expression* condition, Expression* ifTrue, Expression* ifFalse = nullptr) {
     auto* ret = allocator.alloc<If>();
     ret->condition = condition; ret->ifTrue = ifTrue; ret->ifFalse = ifFalse;
     ret->finalize();
     return ret;
   }
+  If* makeIf(Expression* condition, Expression* ifTrue, Expression* ifFalse, WasmType type) {
+    auto* ret = allocator.alloc<If>();
+    ret->condition = condition; ret->ifTrue = ifTrue; ret->ifFalse = ifFalse;
+    ret->finalize(type);
+    return ret;
+  }
   Loop* makeLoop(Name name, Expression* body) {
     auto* ret = allocator.alloc<Loop>();
     ret->name = name; ret->body = body;
@@ -190,6 +225,7 @@ public:
     auto* ret = allocator.alloc<SetGlobal>();
     ret->name = name;
     ret->value = value;
+    ret->finalize();
     return ret;
   }
   Load* makeLoad(unsigned bytes, bool signed_, uint32_t offset, unsigned align, Expression *ptr, WasmType type) {
@@ -199,18 +235,17 @@ public:
     ret->type = type;
     return ret;
   }
-  Load* makeAtomicLoad(unsigned bytes, bool signed_, uint32_t offset, Expression* ptr, WasmType type) {
-    Load* load = makeLoad(bytes, signed_, offset, bytes, ptr, type);
+  Load* makeAtomicLoad(unsigned bytes, uint32_t offset, Expression* ptr, WasmType type) {
+    Load* load = makeLoad(bytes, false, offset, bytes, ptr, type);
     load->isAtomic = true;
     return load;
   }
-  AtomicWait* makeAtomicWait(Expression* ptr, Expression* expected, Expression* timeout, WasmType type) {
+  AtomicWait* makeAtomicWait(Expression* ptr, Expression* expected, Expression* timeout, WasmType expectedType) {
     auto* wait = allocator.alloc<AtomicWait>();
     wait->ptr = ptr;
     wait->expected = expected;
     wait->timeout = timeout;
-    wait->expectedType = type;
-    wait->type = i32;
+    wait->expectedType = expectedType;
     wait->finalize();
     return wait;
   }
@@ -218,7 +253,6 @@ public:
     auto* wake = allocator.alloc<AtomicWake>();
     wake->ptr = ptr;
     wake->wakeCount = wakeCount;
-    wake->type = i32;
     wake->finalize();
     return wake;
   }
@@ -231,7 +265,7 @@ public:
     return ret;
   }
   Store* makeAtomicStore(unsigned bytes, uint32_t offset, Expression* ptr, Expression* value, WasmType type) {
-    Store* store = makeStore(bytes, offset, getWasmTypeSize(type), ptr, value, type);
+    Store* store = makeStore(bytes, offset, bytes, ptr, value, type);
     store->isAtomic = true;
     return store;
   }
@@ -321,15 +355,17 @@ public:
     func->params.push_back(type);
     Index index = func->localNames.size();
     func->localIndices[name] = index;
-    func->localNames.push_back(name);
+    func->localNames[index] = name;
     return index;
   }
 
   static Index addVar(Function* func, Name name, WasmType type) {
     // always ok to add a var, it does not affect other indices
     Index index = func->getNumLocals();
-    if (name.is()) func->localIndices[name] = index;
-    func->localNames.push_back(name);
+    if (name.is()) {
+      func->localIndices[name] = index;
+      func->localNames[index] = name;
+    }
     func->vars.emplace_back(type);
     return index;
   }
@@ -440,6 +476,21 @@ public:
     return makeConst(value);
   }
 
+  // Module-level helpers
+
+  enum Mutability {
+    Mutable,
+    Immutable
+  };
+
+  static Global* makeGlobal(Name name, WasmType type, Expression* init, Mutability mutable_) {
+    auto* glob = new Global;
+    glob->name = name;
+    glob->type = type;
+    glob->init = init;
+    glob->mutable_ = mutable_ == Mutable;
+    return glob;
+  }
 };
 
 } // namespace wasm
