@@ -56,6 +56,13 @@
 
 namespace wasm {
 
+enum Feature : uint32_t {
+  MVP = 0,
+  Atomics = 1 << 0,
+  All = 0xffffffff,
+};
+typedef uint32_t FeatureSet;
+
 // An index in a wasm module
 typedef uint32_t Index;
 
@@ -591,7 +598,7 @@ public:
   Expression* body;
 
   // local names. these are optional.
-  std::vector<Name> localNames;
+  std::map<Index, Name> localNames;
   std::map<Name, Index> localIndices;
 
   struct DebugLocation {
@@ -654,8 +661,7 @@ public:
     Expression* offset;
     std::vector<Name> data;
     Segment() {}
-    Segment(Expression* offset) : offset(offset) {
-    }
+    Segment(Expression* offset) : offset(offset) {}
     Segment(Expression* offset, std::vector<Name>& init) : offset(offset) {
       data.swap(init);
     }
@@ -672,6 +678,7 @@ public:
   Table() : exists(false), imported(false), initial(0), max(kMaxSize) {
     name = Name::fromInt(0);
   }
+  bool hasMax() { return max != kMaxSize; }
 };
 
 class Memory {
@@ -684,6 +691,7 @@ public:
     Expression* offset;
     std::vector<char> data; // TODO: optimize
     Segment() {}
+    Segment(Expression* offset) : offset(offset) {}
     Segment(Expression* offset, const char* init, Address size) : offset(offset) {
       data.resize(size);
       std::copy_n(init, size, data.begin());
@@ -705,6 +713,7 @@ public:
   Memory() : initial(0), max(kMaxSize), exists(false), imported(false), shared(false) {
     name = Name::fromInt(0);
   }
+  bool hasMax() { return max != kMaxSize; }
 };
 
 class Global {
@@ -775,6 +784,7 @@ public:
   void removeImport(Name name);
   void removeExport(Name name);
   void removeFunction(Name name);
+  void removeFunctionType(Name name);
   // TODO: remove* for other elements
 
   void updateMaps();
