@@ -16,7 +16,7 @@
 
 #include "wasm.h"
 #include "wasm-traversal.h"
-#include "ast/branch-utils.h"
+#include "ir/branch-utils.h"
 
 namespace wasm {
 
@@ -390,12 +390,14 @@ void AtomicCmpxchg::finalize() {
 }
 
 void AtomicWait::finalize() {
+  type = i32;
   if (ptr->type == unreachable || expected->type == unreachable || timeout->type == unreachable) {
     type = unreachable;
   }
 }
 
 void AtomicWake::finalize() {
+  type = i32;
   if (ptr->type == unreachable || wakeCount->type == unreachable) {
     type = unreachable;
   }
@@ -578,25 +580,26 @@ bool Function::isVar(Index index) {
 }
 
 bool Function::hasLocalName(Index index) const {
-  return index < localNames.size() && localNames[index].is();
+  return localNames.find(index) != localNames.end();
 }
 
 Name Function::getLocalName(Index index) {
-  assert(hasLocalName(index));
-  return localNames[index];
+  return localNames.at(index);
 }
 
 Name Function::getLocalNameOrDefault(Index index) {
-  if (hasLocalName(index)) {
-    return localNames[index];
+  auto nameIt = localNames.find(index);
+  if (nameIt != localNames.end()) {
+    return nameIt->second;
   }
   // this is an unnamed local
   return Name();
 }
 
 Name Function::getLocalNameOrGeneric(Index index) {
-  if (hasLocalName(index)) {
-    return localNames[index];
+  auto nameIt = localNames.find(index);
+  if (nameIt != localNames.end()) {
+    return nameIt->second;
   }
   return Name::fromInt(index);
 }
@@ -744,7 +747,17 @@ void Module::removeFunction(Name name) {
   functionsMap.erase(name);
 }
 
-  // TODO: remove* for other elements
+void Module::removeFunctionType(Name name) {
+  for (size_t i = 0; i < functionTypes.size(); i++) {
+    if (functionTypes[i]->name == name) {
+      functionTypes.erase(functionTypes.begin() + i);
+      break;
+    }
+  }
+  functionTypesMap.erase(name);
+}
+
+// TODO: remove* for other elements
 
 void Module::updateMaps() {
   functionsMap.clear();
